------------------------------------------------------
-- GFWTooltip.lua
-- Version 2
-- Generic tooltip hooking
-- Based closely on GDI's Reagent Info with additional hooks based on EnhTooltip and RecipeBook... thanks! 
------------------------------------------------------

local THIS_GFWTOOLTIP_VERSION = 2;

if (GFWToolTip_Version == nil or (tonumber(GFWTooltip_Version) ~= nil and GFWTooltip_Version < THIS_GFWTOOLTIP_VERSION)) then
GFWTooltip_Version = THIS_GFWTOOLTIP_VERSION;
-- everything below should only load once

--------------------
-- Function Hooks --
--------------------

local GFWTooltip_OriginalContainerFrameItemButton_OnEnter;
local GFWTooltip_OriginalContainerFrame_Update;
local GFWTooltip_OriginalGameTooltip_SetOwner;
local GFWTooltip_OriginalGameTooltip_OnHide;
local GFWTooltip_OriginalGameTooltip_ClearMoney;
local GFWTooltip_OriginalSetItemRef;
local GFWTooltip_OriginalAuctionFrameItem_OnEnter;
local GFWTooltip_OriginalGameTooltip_SetCraftItem;
local GFWTooltip_OriginalGameTooltip_SetLootItem;
local GFWTooltip_OriginalGameTooltip_SetMerchantItem;
local GFWTooltip_OriginalGameTooltip_SetQuestItem;
local GFWTooltip_OriginalGameTooltip_SetQuestLogItem;
local GFWTooltip_OriginalGameTooltip_SetTradeSkillItem;
local GFWTooltip_OriginalGameTooltip_SetInventoryItem;

-- If non-nil, check for appearance of GameTooltip for adding information
local GFWTooltip_CheckTooltip;

-- If non-nil, don't add extra information to the tooltip on GameTooltip_ClearMoney
local GFWTooltip_SuppressInfoAdd = true;

-- Timer for frequency of tooltip checks
local GFWTooltip_CheckTimer = 0;

-- The current owner of the GameTooltip
local GFWTooltip_GameToolTipOwner;

-- Current Tooltip frame
local GFWTooltip_Tooltip = nil;

-- Boolean to watch if ItemsMatrix is running
local GFWTooltip_ItemsMatrix = false;

-- Hooks for the AllInOneInventory stuff.  Code courtesy of ItemsMatrix
local GFWTooltip_OriginalAllInOneInventory_ModifyItemTooltip;

-- Hooks for the MyInventory stuff.  Code courtesy of ItemsMatrix
local GFWTooltip_OriginalMyInventory_ContainerFrameItemButton_OnEnter;
local GFWTooltip_OriginalMyInventoryFrame_Update;

-- Hook for the LootLink window's tooltip (don't need to hook LL's tooltip function because we already cover everything else it does)
local GFWTooltip_OriginalLootLinkItemButton_OnEnter;

---------------
-- Functions --
---------------

---------------------
-- OnFoo Functions --
---------------------


-- GFWTooltip_OnLoad()
--
-- We call this at the bottom of this file to get things started. 

function GFWTooltip_OnLoad()

	if ( GFWTooltip_Loaded and GFWTooltip_Version >= THIS_GFWTOOLTIP_VERSION ) then return; end -- only load once
	
	-- Hook the GameTooltip's ClearMoney function
	GFWTooltip_OriginalGameTooltip_ClearMoney = GameTooltip_ClearMoney;
	GameTooltip_ClearMoney = GFWTooltip_GameTooltip_ClearMoney;

	-- Hook the GameTooltip's OnHide function
	GFWTooltip_OriginalGameTooltip_OnHide = GameTooltip_OnHide;
	GameTooltip_OnHide = GFWTooltip_GameTooltip_OnHide;

	-- Hook the GameTooltip:SetOwner function
	GFWTooltip_OriginalGameTooltip_SetOwner = GameTooltip.SetOwner;
	GameTooltip.SetOwner = GFWTooltip_GameTooltip_SetOwner;

	-- Hook the SetItemRef function
	GFWTooltip_OriginalSetItemRef = SetItemRef;
	SetItemRef = GFWTooltip_SetItemRef;

	-- Hook ItemsMatrix's functions if they're available
	if (type(ItemsMatrix_AddExtraTooltipInfo) == "function") then
		GFWTooltip_ItemsMatrix = true;
	end

	-- Hook AIOI's stuff if it's available and we're not running ItemsMatrix
	if(AllInOneInventory_Enabled and not GFWTooltip_ItemsMatrix) then
		GFWTooltip_OriginalAllInOneInventory_ModifyItemTooltip = AllInOneInventory_ModifyItemTooltip;
		AllInOneInventory_ModifyItemTooltip = GFWTooltip_AllInOneInventory_ModifyItemTooltip;
	end

	--hook for myinventory
	if (MyInventoryProfile and not GFWTooltip_ItemsMatrix) then
		GFWTooltip_OriginalMyInventory_ContainerFrameItemButton_OnEnter = MyInventory_ContainerFrameItemButton_OnEnter;
		MyInventory_ContainerFrameItemButton_OnEnter = GFWTooltip_MyInventory_ContainerFrameItemButton_OnEnter;

		GFWTooltip_OriginalMyInventoryFrame_Update = MyInventoryFrame_Update;
		MyInventoryFrame_Update = GFWTooltip_MyInventoryFrame_Update;
	end

	-- Hook the LootLink tooltip function
	GFWTooltip_OriginalLootLinkItemButton_OnEnter = LootLinkItemButton_OnEnter;
	LootLinkItemButton_OnEnter = GFWTooltip_LootLinkItemButton_OnEnter;

	-- more hooks for blizzard functions
	GFWTooltip_OriginalContainerFrameItemButton_OnEnter = ContainerFrameItemButton_OnEnter;
	ContainerFrameItemButton_OnEnter = GFWTooltip_ContainerFrameItemButton_OnEnter;
	GFWTooltip_OriginalContainerFrame_Update = ContainerFrame_Update;
	ContainerFrame_Update = GFWTooltip_ContainerFrame_Update;
	GFWTooltip_OriginalAuctionFrameItem_OnEnter = AuctionFrameItem_OnEnter;
	AuctionFrameItem_OnEnter = GFWTooltip_AuctionFrameItem_OnEnter;
	
	GFWTooltip_OriginalGameTooltip_SetLootItem = GameTooltip.SetLootItem;
	GameTooltip.SetLootItem = GFWTooltip_GameTooltip_SetLootItem;
	GFWTooltip_OriginalGameTooltip_SetQuestItem = GameTooltip.SetQuestItem;
	GameTooltip.SetQuestItem = GFWTooltip_GameTooltip_SetQuestItem;
	GFWTooltip_OriginalGameTooltip_SetQuestLogItem = GameTooltip.SetQuestLogItem;
	GameTooltip.SetQuestLogItem = GFWTooltip_GameTooltip_SetQuestLogItem;
	GFWTooltip_OriginalGameTooltip_SetInventoryItem = GameTooltip.SetInventoryItem;
	GameTooltip.SetInventoryItem = GFWTooltip_GameTooltip_SetInventoryItem;
	GFWTooltip_OriginalGameTooltip_SetMerchantItem = GameTooltip.SetMerchantItem;
	GameTooltip.SetMerchantItem = GFWTooltip_GameTooltip_SetMerchantItem;
	GFWTooltip_OriginalGameTooltip_SetCraftItem = GameTooltip.SetCraftItem;
	GameTooltip.SetCraftItem = GFWTooltip_GameTooltip_SetCraftItem;
	GFWTooltip_OriginalGameTooltip_SetTradeSkillItem = GameTooltip.SetTradeSkillItem;
	GameTooltip.SetTradeSkillItem = GFWTooltip_GameTooltip_SetTradeSkillItem;
	GFWTooltip_OriginalGameTooltip_SetAuctionSellItem = GameTooltip.SetAuctionSellItem;
	GameTooltip.SetAuctionSellItem = GFWTooltip_GameTooltip_SetAuctionSellItem;
	
	GFWTooltip_Loaded = true;
	GFWUtils.Print("GFWTooltip loaded.");
end

-- GFWTooltip_OnUpdate(elapsed)
--
-- Ensures that we don't try and write to the tooltips too quickly.

function GFWTooltip_OnUpdate(elapsed)
	GFWTooltip_CheckTimer = GFWTooltip_CheckTimer + elapsed;

	if( GFWTooltip_CheckTimer >= 1 ) then
		if( GFWTooltip_CheckTooltip ) then
			GFWTooltip_CheckTooltip = GFWTooltip_CheckTooltipInfo(GFWTooltip_Tooltip);
		end
		GFWTooltip_CheckTimer = 0;
	end
end

-----------------------
-- Primary Functions --
-----------------------

if (GFWTooltip_Callbacks == nil) then
	GFWTooltip_Callbacks = {};
end

function GFWTooltip_AddCallback(modName, callbackFunction)
	GFWTooltip_Callbacks[modName] = callbackFunction;
end

function GFWTooltip_AddTooltipInfo(frame, name, link, source)

--[[ Sample code for a callback function

	frame:AddLine("test", 1.0, 1.0, 1.0);
	return true;	-- keeps us from modifying it twice; return false or nil if you didn't change the tooltip
]]

	if (frame.gfwDone == 1) then return; end -- we've already been here

	local changedTooltip = false;
	for modName, callback in GFWTooltip_Callbacks do
		if (callback ~= nil and type(callback) == "function") then
			local modifiedInCallback = callback(frame, name, link, source);
			changedTooltip = changedTooltip or modifiedInCallback;
			if (modifiedInCallback) then
				--GFWUtils.Print("Ran tooltip callback for ".. modName..", tooltip modified.");
			else
				--GFWUtils.Print("Ran tooltip callback for ".. modName..", tooltip not modified.");
			end
		end
	end
	--GFWUtils.Print("Done with tooltip callbacks.");
	if (changedTooltip) then
		frame.gfwDone = 1;
		frame:Show();
	end

end

-- Checks the tooltip info for an item name.  If one is found and we haven't updated the tip already, process it.
function GFWTooltip_CheckTooltipInfo(frame, link, source)
	-- If we've already added our information, no need to do it again
	if ( not frame or frame.gfwDone or GFWTooltip_ItemsMatrix) then
		return nil;
	end

	GFWTooltip_Tooltip = frame;
	if (link == nil) then
		link = GFWTooltip_LastLink;
	end
	GFWTooltip_LastLink = link;

	if( frame:IsVisible() ) then
		local field = getglobal(frame:GetName().."TextLeft1");
		if( field and field:IsVisible() ) then
			local name = field:GetText();
			if( name and frame.gfwDone ~= 1) then
				GFWTooltip_AddTooltipInfo(frame, name, link, source);
				return nil;
			end
		end
	end

	return 1;
end

----------------------
-- Helper Functions --
----------------------

-- GFWTooltip_NameFromLink(link)
--
-- Grabs an item's name from its link.  This code is directly from LootLink, so 
-- full props go to Telo.  Well go on, start propping him!

local function GFWTooltip_NameFromLink(link)
	local name;
	if( not link ) then
		return nil;
	end
	for name in string.gfind(link, "|c%x+|Hitem:%d+:%d+:%d+:%d+|h%[(.-)%]|h|r") do
		return name;
	end
	return nil;
end

-- Calling this will allow us to automatically add information to tooltips when needed
function GFWTooltip_AutoInfoOn()
	lSuppressInfoAdd = nil;
end

-- Calling this will prevent us from automatically adding information to tooltips
function GFWTooltip_AutoInfoOff()
	lSuppressInfoAdd = 1;
end

function GFWTooltip_FindItemInBags(findName)
	for bag = 0, 4, 1 do
		size = GetContainerNumSlots(bag);
		if (size) then
			for slot = size, 1, -1 do
				local link = GetContainerItemLink(bag, slot);
				if (link) then
					local name = GFWTooltip_NameFromLink(link);
					if (itemName == findName) then
						return bag, slot;
					end
				end
			end
		end
	end
end


--------------------
-- Hook Functions --
--------------------

-- These are all hooking functions.  They simply replace the built in game functions
-- with our own.  Many are taken directly from Norganna and Telo's code
-- bases.  Nothing to see here.

function GFWTooltip_GameTooltip_SetLootItem(slot)
	GFWTooltip_OriginalGameTooltip_SetLootItem(slot);

	local link = GetLootSlotLink(slot);
	local name = GFWTooltip_NameFromLink(link);
	GFWTooltip_AddTooltipInfo(GameTooltip, name, link, "LOOT");
end

function GFWTooltip_GameTooltip_SetQuestItem(this, qtype, slot)
	GFWTooltip_OriginalGameTooltip_SetQuestItem(this, qtype, slot);

	local link = GetQuestItemLink(qtype, slot);
	local name = GFWTooltip_NameFromLink(link);
	GFWTooltip_AddTooltipInfo(GameTooltip, name, link, "QUEST");
end

function GFWTooltip_GameTooltip_SetQuestLogItem(this, qtype, slot)
	GFWTooltip_OriginalGameTooltip_SetQuestLogItem(this, qtype, slot);

	local link = GetQuestLogItemLink(qtype, slot);
	local name = GFWTooltip_NameFromLink(link);
	GFWTooltip_AddTooltipInfo(GameTooltip, name, link, "QUESTLOG");
end

function GFWTooltip_GameTooltip_SetMerchantItem(this, slot)
	GFWTooltip_OriginalGameTooltip_SetMerchantItem(this, slot);

	local link = GetMerchantItemLink(slot);
	local name = GFWTooltip_NameFromLink(link);
	GFWTooltip_AddTooltipInfo(GameTooltip, name, link, "MERCHANT");
end

function GFWTooltip_AuctionFrameItem_OnEnter(type, index)
	GFWTooltip_OriginalAuctionFrameItem_OnEnter(type, index);

	local link = GetAuctionItemLink(type, index);
	local name = GFWTooltip_NameFromLink(link);
	GFWTooltip_AddTooltipInfo(GameTooltip, name, link, "AUCTION");
end

function GFWTooltip_GameTooltip_SetInventoryItem(this, unit, slot)
	local hasItem, hasCooldown, repairCost = GFWTooltip_OriginalGameTooltip_SetInventoryItem(this, unit, slot);
	local link = GetInventoryItemLink(unit, slot);
	local name = GFWTooltip_NameFromLink(link);
	GFWTooltip_AddTooltipInfo(GameTooltip, name, link, "INVENTORY");

	return hasItem, hasCooldown, repairCost;
end

function GFWTooltip_GameTooltip_SetCraftItem(this, skill, slot)
	GFWTooltip_OriginalGameTooltip_SetCraftItem(this, skill, slot);
	local link;
	if (slot) then
		link = GetCraftReagentItemLink(skill, slot);
		local name = GFWTooltip_NameFromLink(link);
		GFWTooltip_AddTooltipInfo(GameTooltip, name, link, "CRAFT_REAGENT");
	else
		link = GetCraftItemLink(skill);
		local name = GFWTooltip_NameFromLink(link);
		GFWTooltip_AddTooltipInfo(GameTooltip, name, link, "CRAFT_ITEM");
	end
end

function GFWTooltip_GameTooltip_SetTradeSkillItem(this, skill, slot)
	GFWTooltip_OriginalGameTooltip_SetTradeSkillItem(this, skill, slot);
	local link;
	if (slot) then
		link = GetTradeSkillReagentItemLink(skill, slot);
		local name = GFWTooltip_NameFromLink(link);
		GFWTooltip_AddTooltipInfo(GameTooltip, name, link, "TRADESKILL_REAGENT");
	else
		link = GetTradeSkillItemLink(skill);
		local name = GFWTooltip_NameFromLink(link);
		GFWTooltip_AddTooltipInfo(GameTooltip, name, link, "TRADESKILL_ITEM");
	end
end

function GFWTooltip_GameTooltip_SetAuctionSellItem(this)
	GFWTooltip_OriginalGameTooltip_SetAuctionSellItem(this);
    local name, texture, quantity, quality, canUse, price = GetAuctionSellItemInfo();
	if (name) then
		local bag, slot = GFWTooltip_FindItemInBags(name);
		if (bag) then
			local link = GetContainerItemLink(bag, slot);
			GFWTooltip_AddTooltipInfo(GameTooltip, name, link, "AUCTION_SELL");
		end
	end
end

function GFWTooltip_ContainerFrameItemButton_OnEnter()
	GFWTooltip_OriginalContainerFrameItemButton_OnEnter();
	GFWTooltip_AutoInfoOff();

	if (GFWTooltip_ItemsMatrix) then
		GFWTooltip_AutoInfoOn();
		return;
	end

	if (not InRepairMode()) then
		local frameID = this:GetParent():GetID();
		local buttonID = this:GetID();
		local link = GetContainerItemLink(frameID, buttonID);
		local name = GFWTooltip_NameFromLink(link);

		if( name ) then
			local texture, itemCount, locked, quality, readable = GetContainerItemInfo(frameID, buttonID);
			GFWTooltip_CheckTooltipInfo(GameTooltip, link, "CONTAINER");
			GameTooltip:Show();
		end
	end
	GFWTooltip_AutoInfoOn();
end

function GFWTooltip_ContainerFrame_Update(frame)
	GFWTooltip_OriginalContainerFrame_Update(frame);
	GFWTooltip_AutoInfoOff();

	if (GFWTooltip_ItemsMatrix) then
		GFWTooltip_AutoInfoOn();
		return;
	end

	if( not InRepairMode() and GameTooltip:IsVisible() ) then
		local frameID = frame:GetID();
		local frameName = frame:GetName();
		local iButton;
		for iButton = 1, frame.size do
			local button = getglobal(frameName.."Item"..iButton);
			if( GameTooltip:IsOwned(button) ) then
				local buttonID = button:GetID();
				local link = GetContainerItemLink(frameID, buttonID);
				local name = GFWTooltip_NameFromLink(link);

				if( name ) then
					local texture, itemCount, locked, quality, readable = GetContainerItemInfo(frameID, buttonID);
					GFWTooltip_CheckTooltipInfo(GameTooltip, link, "CONTAINER");
					GameTooltip:Show();
				end
			end
		end
	end
	GFWTooltip_AutoInfoOn();
end

function GFWTooltip_GameTooltip_ClearMoney()
	GFWTooltip_OriginalGameTooltip_ClearMoney();
	if (not GFWTooltip_SuppressInfoAdd) then
		GFWTooltip_CheckTooltip = GFWTooltip_CheckTooltipInfo(GameTooltip, nil, "MONEY");
	end
end

function GFWTooltip_GameTooltip_OnHide()
	GFWTooltip_OriginalGameTooltip_OnHide();
	GameTooltip.gfwDone = nil;
	if ( GFWTooltip_Tooltip ) then
		GFWTooltip_Tooltip.gfwDone = nil;
		GFWTooltip_Tooltip = nil;
	end
end

function GFWTooltip_GameTooltip_SetOwner(this, owner, anchor)
	GFWTooltip_OriginalGameTooltip_SetOwner(this, owner, anchor);
	GFWTooltip_GameToolTipOwner = owner;
end

function GFWTooltip_SetItemRef(link, text, button)
	GFWTooltip_OriginalSetItemRef(link, text, button);
	GFWTooltip_CheckTooltip = GFWTooltip_CheckTooltipInfo(ItemRefTooltip, link, "ITEMREF");
end

-- AIOI Hooks
function GFWTooltip_AllInOneInventory_ModifyItemTooltip(bag, slot, tooltipName)
	GFWTooltip_OriginalAllInOneInventory_ModifyItemTooltip(bag, slot, tooltipName);

	-- Verify AIOI is installed and running
	if(not AllInOneInventory_Enabled or GFWTooltip_ItemsMatrix) then
		return;	
	end

	local tooltip = getglobal(tooltipName);
	if ( not tooltip ) then
		tooltip = getglobal("GameTooltip");
		tooltipName = "GameTooltip";
	end
	if ( not tooltip ) then
		return false;
	end

	if ( not InRepairMode() ) then
		local link = GetContainerItemLink(bag, slot);
		local name = GFWTooltip_NameFromLink(link);

		GFWTooltip_AddTooltipInfo(frame, name, link, "AIOI");
	end
end

-- MyInventory Hooks
function GFWTooltip_MyInventory_ContainerFrameItemButton_OnEnter()
	GFWTooltip_OriginalMyInventory_ContainerFrameItemButton_OnEnter();

	if (GameTooltip:IsVisible() and not InRepairMode()) then
		local bag, slot = MyInventory_GetIdAsBagSlot(this:GetID());
		local _, stack = GetContainerItemInfo(bag, slot);

		local link = GetContainerItemLink(bag, slot);
		local name = GFWTooltip_NameFromLink(link);

		if(link and name) then
			GFWTooltip_AddTooltipInfo(frame, name, link, "MYINV");
		end
	end
end

function GFWTooltip_MyInventoryFrame_Update(frame)
	GFWTooltip_OriginalMyInventoryFrame_Update(frame);

	local name = frame:GetName();
	for j=1, frame.size, 1 do
		local itemButton = getglobal(name.."Item"..j);

		if (GameTooltip:IsVisible() and GameTooltip:IsOwned(itemButton)) then
			tooltip = getglobal("GameTooltip");
			tooltipName = "GameTooltip";

			if ( not tooltip ) then
				return false;
			end

			local bag, slot = MyInventory_GetIdAsBagSlot(itemButton:GetID());
			local link = GetContainerItemLink(bag, slot);
			local name = GFWTooltip_NameFromLink(link);

			if (name ~= nil) then
				GFWTooltip_AddTooltipInfo(GameTooltip, name, link, "MYINV");
			end
		end
	end
end

function GFWTooltip_LootLinkItemButton_OnEnter()
	GFWTooltip_OriginalLootLinkItemButton_OnEnter();
	local name = this:GetText();
	local itemLink = ItemLinks[name];
	local link;
	if (itemLink and itemLink.c and itemLink.i and LootLink_CheckItemServer(itemLink, LootLinkState.ServerNamesToIndices[GetCVar("realmName")])) then
		local item = string.gsub(itemLink.i, "(%d+):(%d+):(%d+):(%d+)", "%1:0:%3:%4");
		link = "|c"..itemLink.c.."|Hitem:"..item.."|h["..name.."]|h|r";
	end
	if (link) then
		GFWTooltip_AddTooltipInfo(LootLinkTooltip, name, link, "LOOTLINK");
	end
end

--------------------
-- Loading
--------------------

GFWTooltip_OnLoad();

end -- for initial version check
